use short_uuid::CustomTranslator;
use std::collections::HashMap;
use std::hash::{DefaultHasher, Hash, Hasher};

/// Memos is written in go and use https://github.com/lithammer/shortuuid for uid generation.
/// Considering that these are just uuids, we can get an u128 from them.
struct ShortUuidConverter {
    translator: CustomTranslator,
}

impl ShortUuidConverter {
    /// The go library uses a custom alphabet, so we need to use that too.
    /// Taken from https://github.com/lithammer/shortuuid/blob/ffec47be5937285a5ee628d94b2dcbb8cb58e22d/alphabet.go#L12
    const SHORT_UID_ALPHA: &'static str =
        "23456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz";

    fn new() -> Self {
        Self {
            translator: CustomTranslator::new(Self::SHORT_UID_ALPHA).unwrap(),
        }
    }

    fn try_to_u128(&self, input: &str) -> anyhow::Result<u128> {
        let short = match short_uuid::ShortUuidCustom::parse_str(input, &self.translator) {
            Ok(short) => short,
            Err(reason) => return Err(anyhow::anyhow!("Failed to parse {}: {}", input, reason)),
        };
        let uid = short.to_uuid(&self.translator).unwrap();
        Ok(uid.as_u128())
    }

    fn from_u128(&self, input: u128) -> String {
        let uid = uuid::Uuid::from_u128(input);
        let short = short_uuid::ShortUuidCustom::from_uuid(&uid, &self.translator);
        short.to_string()
    }
}

pub struct INodeManager {
    nodes: HashMap<u64, u128>,
    short_uuid_converter: ShortUuidConverter,
}

impl INodeManager {
    pub fn new() -> Self {
        Self {
            nodes: HashMap::new(),
            short_uuid_converter: ShortUuidConverter::new(),
        }
    }

    pub fn get_or_create_inode(&mut self, uid: &str) -> u64 {
        let num = self.short_uuid_converter.try_to_u128(uid).unwrap();
        let inode = self.hash_num(num);
        self.nodes.insert(inode, num);
        inode
    }

    pub fn get_uid(&self, inode: u64) -> Option<String> {
        match self.nodes.get(&inode).copied() {
            None => None,
            Some(num) => Some(self.short_uuid_converter.from_u128(num)),
        }
    }

    fn hash_num(&self, num: u128) -> u64 {
        let mut hasher = DefaultHasher::new();
        num.hash(&mut hasher);
        hasher.finish()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    /// (original_uid, expected_uuid_u128, expected_inode)
    const TEST_IDS: [(&str, u128, u64); 10] = [
        (
            "VgnhpfGXw7vHWPce6oNaVS",
            206797227597459323123527397586500652966,
            3913517761545603567,
        ),
        (
            "hqKG23Ds8hiZXziKorZvHH",
            297564317390508405213354428528166123258,
            10895052826299621117,
        ),
        (
            "JBSRgn6FcAP3s3S8WHC6Wy",
            120769892440230658418002098667456834855,
            925797026661051322,
        ),
        (
            "Bp7Xgaz3rs6F5e9SddHLcP",
            73279780794183508306623012303071573694,
            16832634417085068821,
        ),
        (
            "ELoMHpye6sqJKceTTZmyCH",
            92114081947781017664946393632646945141,
            3318364556962759386,
        ),
        (
            "7CppGL2gXD7meWV2GLFnLZ",
            38772739364293647929273214809052070870,
            17639521134799251273,
        ),
        (
            "PKK3i89cZZ3ELGjER85KND",
            159155872788710468840620268141821518047,
            5604479176214409706,
        ),
        (
            "bpdrRf6Y4w8X9BAL4TJZV6",
            252650725998566377086795281031975766693,
            6088146008880505408,
        ),
        (
            "SHtDbXtL7M7iyTnSQaMyrk",
            181382667582346113605641996269296312097,
            15579490772996943984,
        ),
        (
            "GA8XqyXDNFEko6nW5UFbHh",
            105655915756677130982417952151348309032,
            5472690243812647263,
        ),
    ];

    #[test]
    fn convert_short_uids() {
        let converter = ShortUuidConverter::new();
        for (input, expected, _) in TEST_IDS {
            let num = converter.try_to_u128(input).unwrap();
            assert_eq!(num, expected, "Failed to decode {}", input);
            let enc = converter.from_u128(num);
            assert_eq!(enc, input, "Failed to encode {}", num);
        }
    }

    #[test]
    fn inode_converter() {
        let mut inode_manager = INodeManager::new();
        for (input, _, expected) in TEST_IDS {
            let num = inode_manager.get_or_create_inode(input);
            assert_eq!(num, expected, "Failed to decode {}", input);
            let enc = inode_manager.get_uid(num);
            assert_eq!(enc.unwrap(), input, "Failed to encode {}", num);
        }
    }
}
