mod inodes;

use fuser::{
    FileAttr, FileType, Filesystem, MountOption, ReplyAttr, ReplyData, ReplyDirectory, ReplyEntry,
    Request,
};
use nix::libc::{EIO, ENOENT};
use reqwest::Method;
use serde::Deserialize;
use std::ffi::OsStr;
use std::time::{Duration, UNIX_EPOCH};

const BASE_TTL: Duration = Duration::from_secs(30);

#[derive(Debug, Clone, Deserialize)]
pub struct MemoResponse {
    pub memos: Vec<Memo>,
}

#[derive(Debug, Clone, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Memo {
    pub name: String,
    pub state: String,
    pub creator: String,
    pub create_time: chrono::DateTime<chrono::Utc>,
    pub update_time: Option<chrono::DateTime<chrono::Utc>>,
    pub display_time: Option<chrono::DateTime<chrono::Utc>>,
    pub content: String,
    pub visibility: String,
    pub tags: Vec<String>,
    pub snippet: String,
    pub relations: Vec<MemoRelation>,
}

impl Memo {
    pub fn into_file_attr(self, ino: u64, uid: u32, gid: u32) -> FileAttr {
        let len = self.content.len() as u64;
        FileAttr {
            ino,
            size: len,
            blocks: f64::ceil(len as f64 / 512.0) as u64,
            atime: self.display_time.unwrap_or(self.create_time).into(),
            mtime: self.update_time.unwrap_or(self.create_time).into(),
            ctime: self.update_time.unwrap_or(self.create_time).into(),
            crtime: self.create_time.into(),
            kind: FileType::RegularFile,
            perm: 0o644,
            nlink: 1,
            uid,
            gid,
            rdev: 0,
            flags: 0,
            blksize: 512,
        }
    }
}

#[derive(Debug, Clone, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct MemoRelation {
    memo: MemoRelationMemo,
    related_memo: MemoRelationMemo,
    #[serde(rename = "type")]
    relation_type: String,
}

#[derive(Debug, Clone, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct MemoRelationMemo {
    name: String,
    uid: String,
    snippet: String,
}

fn get_dir_attr(ino: u64, uid: u32, gid: u32) -> FileAttr {
    FileAttr {
        ino,
        size: 0,
        blocks: 0,
        atime: UNIX_EPOCH,
        mtime: UNIX_EPOCH,
        ctime: UNIX_EPOCH,
        crtime: UNIX_EPOCH,
        kind: FileType::Directory,
        perm: 0o755,
        nlink: 2,
        uid,
        gid,
        rdev: 0,
        flags: 0,
        blksize: 512,
    }
}

struct MemosFS {
    api_base: String,
    token: String,
    inode_manager: inodes::INodeManager,
    cache: moka::sync::Cache<String, Memo>,
    all_cache: moka::sync::Cache<String, MemoResponse>,
    uid: u32,
    gid: u32,
}

impl MemosFS {
    fn new(api_base: &str, token: &str) -> Self {
        MemosFS {
            api_base: api_base.to_string(),
            token: token.to_string(),
            inode_manager: inodes::INodeManager::new(),
            cache: moka::sync::CacheBuilder::new(500)
                .time_to_live(Duration::from_secs(30))
                .build(),
            all_cache: moka::sync::CacheBuilder::new(5)
                .time_to_live(Duration::from_secs(30))
                .build(),
            uid: nix::unistd::geteuid().as_raw(),
            gid: nix::unistd::getegid().as_raw(),
        }
    }

    fn get_memo_from_id(&self, id: &str) -> Option<Memo> {
        if let Some(memo) = self.cache.get(id) {
            return Some(memo);
        }

        println!("fetching {}", id);

        let url = format!("{}/v1/memos/{}", self.api_base, id);
        let mut req =
            reqwest::blocking::Request::new(Method::GET, reqwest::Url::parse(&url).unwrap());
        req.headers_mut().insert(
            "Authorization",
            format!("Bearer {}", self.token).parse().unwrap(),
        );
        let res = reqwest::blocking::Client::new().execute(req).unwrap();

        match res.json::<Memo>() {
            Ok(memo) => {
                self.cache.insert(id.to_string(), memo.clone());
                Some(memo)
            }
            Err(reason) => {
                eprintln!("Failed to parse memo: {}", reason);
                None
            }
        }
    }
}

impl Filesystem for MemosFS {
    fn lookup(&mut self, _req: &Request, parent: u64, name: &OsStr, reply: ReplyEntry) {
        if parent != 1 {
            reply.error(ENOENT);
            return;
        }

        let name = match name.to_str() {
            Some(x) => x,
            None => {
                reply.error(ENOENT);
                return;
            }
        };

        if name.len() != 25 {
            reply.error(ENOENT);
            return;
        }

        let name_without_ext = &name[0..22];

        let ino = self.inode_manager.get_or_create_inode(name_without_ext);

        match self.get_memo_from_id(&name_without_ext) {
            Some(memo) => {
                let attr = memo.into_file_attr(ino, self.uid, self.gid);
                reply.entry(&BASE_TTL, &attr, 0)
            }
            None => {
                reply.error(ENOENT);
            }
        }
    }

    fn getattr(&mut self, _req: &Request, ino: u64, _fh: Option<u64>, reply: ReplyAttr) {
        if ino == 1 {
            reply.attr(&BASE_TTL, &get_dir_attr(1, self.uid, self.gid));
            return;
        }

        let id = match self.inode_manager.get_uid(ino) {
            Some(id) => id,
            None => {
                reply.error(ENOENT);
                return;
            }
        };

        match self.get_memo_from_id(&id) {
            Some(memo) => {
                let attr = memo.into_file_attr(ino, self.uid, self.gid);
                reply.attr(&BASE_TTL, &attr)
            }
            None => {
                reply.error(ENOENT);
            }
        }
    }

    fn read(
        &mut self,
        _req: &Request,
        ino: u64,
        _fh: u64,
        offset: i64,
        _size: u32,
        _flags: i32,
        _lock: Option<u64>,
        reply: ReplyData,
    ) {
        let id = match self.inode_manager.get_uid(ino) {
            Some(id) => id,
            None => {
                reply.error(ENOENT);
                return;
            }
        };

        match self.get_memo_from_id(&id) {
            Some(memo) => {
                reply.data(&memo.content.as_bytes()[offset as usize..]);
            }
            None => {
                reply.error(ENOENT);
            }
        }
    }

    fn readdir(
        &mut self,
        _req: &Request,
        ino: u64,
        _fh: u64,
        offset: i64,
        mut reply: ReplyDirectory,
    ) {
        if ino != 1 {
            reply.error(ENOENT);
            return;
        }

        if offset == 0 {
            if reply.add(1, 1, FileType::Directory, ".") {
                reply.ok();
                return;
            }
        }

        if offset <= 1 {
            if reply.add(1, 2, FileType::Directory, "..") {
                reply.ok();
                return;
            }
        }

        if offset >= 2 {
            let entries = match self.all_cache.get("all") {
                Some(all) => all,
                None => {
                    println!("Fetching all memos");
                    let uri = format!("{}/v1/memos?page_size=100", self.api_base);
                    let mut req = reqwest::blocking::Request::new(
                        Method::GET,
                        reqwest::Url::parse(&uri).unwrap(),
                    );
                    req.headers_mut().insert(
                        "Authorization",
                        format!("Bearer {}", self.token).parse().unwrap(),
                    );

                    let res = match reqwest::blocking::Client::new().execute(req) {
                        Ok(res) => res,
                        Err(_) => {
                            reply.error(EIO);
                            return;
                        }
                    };

                    let entries = match res.json::<MemoResponse>() {
                        Ok(entries) => entries,
                        Err(_) => {
                            reply.error(EIO);
                            return;
                        }
                    };

                    self.all_cache.insert("all".to_string(), entries.clone());

                    entries
                }
            };

            for (memo_index, entry) in entries.memos.into_iter().enumerate() {
                let current_offset = (memo_index + 2) as i64;

                if current_offset < offset {
                    continue;
                }

                let name = entry.name.split('/').last().unwrap();
                let inode = self.inode_manager.get_or_create_inode(name);
                let next_offset = current_offset + 1;

                self.cache.insert(name.to_string(), entry.clone());

                if reply.add(
                    inode,
                    next_offset,
                    FileType::RegularFile,
                    format!("{}.md", name),
                ) {
                    break;
                }
            }
        }

        reply.ok();
    }
}

fn main() {
    let args: Vec<String> = std::env::args().collect();
    if args.len() != 4 {
        eprintln!("Usage: {} <mountpoint> <api_base> <access_token>", args[0]);
        std::process::exit(1);
    }

    let mountpoint = &args[1];
    let api_base = &args[2];
    let access_token = &args[3];
    let options = vec![MountOption::RO, MountOption::FSName("memosfs".to_string())];

    fuser::mount2(MemosFS::new(api_base, access_token), mountpoint, &options).unwrap();
}
